@@ -29,7 +29,7 @@
 #include <numa.h>
 #include <math.h>
 #include <limits.h>
-#include <linux/unistd.h>
+#include <inttypes.h>
 
 #include <sys/prctl.h>
 #include <sys/stat.h>
@@ -41,9 +41,11 @@
 #include <sys/mman.h>
 #include <sys/syscall.h>
 
+#include <linux/unistd.h>
+
 #include "rt-utils.h"
 #include "rt-numa.h"
-#include "error.h"
+#include "rt-error.h"
 
 #ifdef __GNUC__
 # define atomic_inc(ptr)   __sync_add_and_fetch((ptr), 1)
@@ -169,8 +171,10 @@ struct global {
 	uint64_t              workload_mem_size;
 	int                   enable_bias;
 	uint64_t              bias;
+	int                   quiet;
 	int                   single_preheat_thread;
 	int                   output_omit_zero_buckets;
+	char                  jsonfile[MAX_PATH];
 
 	/* Mutable state. */
 	volatile enum command cmd;
@@ -479,6 +483,39 @@ static void write_summary(struct thread *t)
 	printf("\n");
 }
 
+static void write_summary_json(FILE *f, void *data)
+{
+	struct thread *t = data;
+	int i, j, comma;
+
+	fprintf(f, "  \"num_threads\": %d,\n", g.n_threads);
+	fprintf(f, "  \"thread\": {\n");
+	for (i = 0; i < g.n_threads; ++i) {
+		fprintf(f, "    \"%u\": {\n", i);
+		fprintf(f, "      \"cpu\": %d,\n", t[i].core_i);
+		fprintf(f, "      \"freq\": %d,\n", t[i].cpu_mhz);
+		fprintf(f, "      \"min\": %" PRIu64 ",\n", t[i].minlat);
+		fprintf(f, "      \"avg\": %3lf,\n", t[i].average);
+		fprintf(f, "      \"max\": %" PRIu64 ",\n", t[i].maxlat);
+		fprintf(f, "      \"duration\": %.3f,\n",
+			cycles_to_sec(&(t[i]), t[i].runtime));
+		fprintf(f, "      \"histogram\": {");
+		for (j = 0, comma = 0; j < g.bucket_size; j++) {
+			if (t[i].buckets[j] == 0)
+				continue;
+			fprintf(f, "%s", comma ? ",\n" : "\n");
+			fprintf(f, "        \"%" PRIu64 "\": %" PRIu64,
+				g.bias+j+1, t[i].buckets[j]);
+			comma = 1;
+		}
+		if (comma)
+			fprintf(f, "\n");
+		fprintf(f, "      }\n");
+		fprintf(f, "    }%s\n", i == g.n_threads - 1 ? "" : ",");
+	}
+	fprintf(f, "  }\n");
+}
+
 static void run_expt(struct thread *threads, int runtime_secs)
 {
 	int i;
@@ -512,7 +549,6 @@ static void handle_alarm(int code)
 
 static void usage(int error)
 {
-	printf("oslat V %1.2f\n", VERSION);
 	printf("Usage:\n"
 	       "oslat <options>\n\n"
 	       "This is an OS latency detector by running busy loops on specified cores.\n"
@@ -524,11 +560,13 @@ static void usage(int error)
 	       "-C, --cpu-main-thread  Specify which CPU the main thread runs on.  Default is cpu0.\n"
 	       "-D, --duration         Specify test duration, e.g., 60, 20m, 2H\n"
 	       "                       (m/M: minutes, h/H: hours, d/D: days)\n"
+	       "    --json=FILENAME    write final results into FILENAME, JSON formatted\n"
 	       "-f, --rtprio           Using SCHED_FIFO priority (1-99)\n"
 	       "-m, --workload-mem     Size of the memory to use for the workload (e.g., 4K, 1M).\n"
 	       "                       Total memory usage will be this value multiplies 2*N,\n"
 	       "                       because there will be src/dst buffers for each thread, and\n"
 	       "                       N is the number of processors for testing.\n"
+	       "-q  --quiet            print a summary only on exit\n"
 	       "-s, --single-preheat   Use a single thread when measuring latency at preheat stage\n"
 	       "                       NOTE: please make sure the CPU frequency on all testing cores\n"
 	       "                       are locked before using this parmater.  If you don't know how\n"
@@ -557,34 +595,46 @@ static int workload_select(char *name)
 	return -1;
 }
 
+enum option_value {
+	OPT_BUCKETSIZE=1, OPT_CPU_LIST, OPT_CPU_MAIN_THREAD,
+	OPT_DURATION, OPT_JSON, OPT_RT_PRIO, OPT_HELP, OPT_TRACE_TH,
+	OPT_WORKLOAD, OPT_WORKLOAD_MEM, OPT_BIAS,
+	OPT_QUIET, OPT_SINGLE_PREHEAT, OPT_ZERO_OMIT,
+	OPT_VERSION
+};
+
 /* Process commandline options */
 static void parse_options(int argc, char *argv[])
 {
 	while (1) {
+		int option_index = 0;
 		static struct option options[] = {
-			{ "bucket-size", required_argument, NULL, 'b' },
-			{ "cpu-list", required_argument, NULL, 'c' },
-			{ "cpu-main-thread", required_argument, NULL, 'C'},
-			{ "duration", required_argument, NULL, 'D' },
-			{ "rtprio", required_argument, NULL, 'f' },
-			{ "help", no_argument, NULL, 'h' },
-			{ "trace-threshold", required_argument, NULL, 'T' },
-			{ "workload", required_argument, NULL, 'w'},
-			{ "workload-mem", required_argument, NULL, 'm'},
-			{ "bias", no_argument, NULL, 'B'},
-			{ "single-preheat", no_argument, NULL, 's'},
-			{ "zero-omit", no_argument, NULL, 'u'},
-			{ "version", no_argument, NULL, 'v'},
+			{ "bucket-size", required_argument,	NULL, OPT_BUCKETSIZE },
+			{ "cpu-list",	required_argument,	NULL, OPT_CPU_LIST },
+			{ "cpu-main-thread", required_argument, NULL, OPT_CPU_MAIN_THREAD},
+			{ "duration",	required_argument,	NULL, OPT_DURATION },
+			{ "json",	required_argument,      NULL, OPT_JSON },
+			{ "rtprio",	required_argument,	NULL, OPT_RT_PRIO },
+			{ "help",	no_argument,		NULL, OPT_HELP },
+			{ "trace-threshold", required_argument,	NULL, OPT_TRACE_TH },
+			{ "workload",	required_argument,	NULL, OPT_WORKLOAD },
+			{ "workload-mem", required_argument,	NULL, OPT_WORKLOAD_MEM },
+			{ "bias",	no_argument,		NULL, OPT_BIAS },
+			{ "quiet",	no_argument,		NULL, OPT_QUIET },
+			{ "single-preheat", no_argument,	NULL, OPT_SINGLE_PREHEAT },
+			{ "zero-omit",	no_argument,		NULL, OPT_ZERO_OMIT },
+			{ "version",	no_argument,		NULL, OPT_VERSION },
 			{ NULL, 0, NULL, 0 },
 		};
-		int i, c = getopt_long(argc, argv, "b:Bc:C:D:f:hm:sw:T:vz",
-				       options, NULL);
+		int i, c = getopt_long(argc, argv, "b:Bc:C:D:f:hm:qsw:T:vz",
+				       options, &option_index);
 		long ncores;
 
 		if (c == -1)
 			break;
 
 		switch (c) {
+		case OPT_BUCKETSIZE:
 		case 'b':
 			g.bucket_size = strtol(optarg, NULL, 10);
 			if (g.bucket_size > 1024 || g.bucket_size <= 4) {
@@ -593,12 +643,15 @@ static void parse_options(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case OPT_BIAS:
 		case 'B':
 			g.enable_bias = 1;
 			break;
+		case OPT_CPU_LIST:
 		case 'c':
 			g.cpu_list = strdup(optarg);
 			break;
+		case OPT_CPU_MAIN_THREAD:
 		case 'C':
 			ncores = sysconf(_SC_NPROCESSORS_CONF);
 			g.cpu_main_thread = strtol(optarg, NULL, 10);
@@ -608,6 +661,7 @@ static void parse_options(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case OPT_DURATION:
 		case 'D':
 			g.runtime = parse_time_string(optarg);
 			if (!g.runtime) {
@@ -615,6 +669,7 @@ static void parse_options(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case OPT_RT_PRIO:
 		case 'f':
 			g.rtprio = strtol(optarg, NULL, 10);
 			if (g.rtprio < 1 || g.rtprio > 99) {
@@ -622,6 +677,10 @@ static void parse_options(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case OPT_JSON:
+			strncpy(g.jsonfile, optarg, strnlen(optarg, MAX_PATH-1));
+			break;
+		case OPT_TRACE_TH:
 		case 'T':
 			g.trace_threshold = strtol(optarg, NULL, 10);
 			if (g.trace_threshold <= 0) {
@@ -630,6 +689,7 @@ static void parse_options(int argc, char *argv[])
 			}
 			enable_trace_mark();
 			break;
+		case OPT_WORKLOAD:
 		case 'w':
 			if (workload_select(optarg)) {
 				printf("Unknown workload '%s'.  Please choose from: ", optarg);
@@ -642,12 +702,18 @@ static void parse_options(int argc, char *argv[])
 				exit(1);
 			}
 			break;
+		case OPT_WORKLOAD_MEM:
 		case 'm':
 			if (parse_mem_string(optarg, &g.workload_mem_size)) {
 				printf("Unknown workload memory size '%s'.\n\n", optarg);
 				exit(1);
 			}
 			break;
+		case OPT_QUIET:
+		case 'q':
+			g.quiet = 1;
+			break;
+		case OPT_SINGLE_PREHEAT:
 		case 's':
 			/*
 			 * Only use one core for pre-heat.  Then if --bias is used, the
@@ -655,16 +721,19 @@ static void parse_options(int argc, char *argv[])
 			 */
 			g.single_preheat_thread = true;
 			break;
+		case OPT_VERSION:
 		case 'v':
 			/*
-			 * Because we always dump the version even before parsing options,
-			 * what we need to do is to quit..
+			 * We always print the version before parsing options
+			 * so just exit
 			 */
 			exit(0);
 			break;
+		case OPT_ZERO_OMIT:
 		case 'z':
 			g.output_omit_zero_buckets = 1;
 			break;
+		case OPT_HELP:
 		case 'h':
 			usage(0);
 			break;
@@ -716,7 +785,6 @@ int main(int argc, char *argv[])
 	struct thread *threads;
 	int i, n_cores;
 	struct bitmask *cpu_set = NULL;
-	int max_cpus = sysconf(_SC_NPROCESSORS_ONLN);
 
 #ifdef FRC_MISSING
 	printf("This architecture is not yet supported. "
@@ -728,6 +796,8 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	rt_init(argc, argv);
+
 	g.app_name = argv[0];
 	g.rtprio = 0;
 	g.bucket_size = BUCKET_SIZE;
@@ -736,21 +806,26 @@ int main(int argc, char *argv[])
 	g.workload_mem_size = WORKLOAD_MEM_SIZE;
 	/* Run the main thread on cpu0 by default */
 	g.cpu_main_thread = 0;
-
+	printf("oslat V %1.2f\n", VERSION);
 	parse_options(argc, argv);
 
 	TEST(mlockall(MCL_CURRENT | MCL_FUTURE) == 0);
 
 	if (!g.cpu_list)
 		g.cpu_list = strdup("all");
-	if (parse_cpumask(g.cpu_list, max_cpus, &cpu_set))
-		exit(1);
+
+	cpu_set = numa_parse_cpustring_all(g.cpu_list);
+	if (!cpu_set)
+		fatal("oslat: parse_cpumask failed.\n");
 	n_cores = numa_bitmask_weight(cpu_set);
 
 	TEST(threads = calloc(1, n_cores * sizeof(threads[0])));
-	for (i = 0; i < n_cores; ++i)
-		if (numa_bitmask_isbitset(cpu_set, i) && move_to_core(i) == 0)
+	for (i = 0; n_cores && i < cpu_set->size; i++) {
+		if (numa_bitmask_isbitset(cpu_set, i) && move_to_core(i) == 0) {
 			threads[g.n_threads_total++].core_i = i;
+			n_cores--;
+		}
+	}
 
 	if (numa_bitmask_isbitset(cpu_set, 0) && g.rtprio)
 		printf("WARNING: Running SCHED_FIFO workload on CPU 0 may hang the thread\n");
@@ -763,9 +838,11 @@ int main(int argc, char *argv[])
 	signal(SIGINT, handle_alarm);
 	signal(SIGTERM, handle_alarm);
 
-	dump_globals();
+	if (!g.quiet)
+		dump_globals();
 
-	printf("Pre-heat for 1 seconds...\n");
+	if (!g.quiet)
+		printf("Pre-heat for 1 seconds...\n");
 	if (g.single_preheat_thread)
 		g.n_threads = 1;
 	else
@@ -773,15 +850,20 @@ int main(int argc, char *argv[])
 	run_expt(threads, 1);
 	record_bias(threads);
 
-	printf("Test starts...\n");
+	if (!g.quiet)
+		printf("Test starts...\n");
 	/* Reset n_threads to always run on all the cores */
 	g.n_threads = g.n_threads_total;
 	run_expt(threads, g.runtime);
 
-	printf("Test completed.\n\n");
+	if (!g.quiet)
+		printf("Test completed.\n\n");
 
 	write_summary(threads);
 
+	if (strlen(g.jsonfile) != 0)
+		rt_write_json(g.jsonfile, 0, write_summary_json, threads);
+
 	if (g.cpu_list) {
 		free(g.cpu_list);
 		g.cpu_list = NULL;
